import numpy as np
from numpy import sin, cos, tan, pi, exp, sqrt
from pprint import pprint

from tklib.tkobject import tkObject
import tklib.tkre as tkre
from tklib.tksci.tkoptimizeobject import tkOptimizeData
from tklib.tkutils import lvlprint, merge_attributes, joinf



"""
August 14, 2019: Transported to python3 by Toshio Kamiya from:
服部力、名取亮、小国力 監修
Fortranによる数値計算ソフトウェア （フロッピーディスク付き）
丸善株式会社、平成元年（１９８９年）１２月３０日発行
"""


class tkSimplexVertex(tkObject):
    def __init__(self, nvar, **args):
        self.nvar = nvar
        self.x = []
        for i in range(self.nvar):
            self.x.append(0.0)
        self.f = None
        self.update(**args)

    def __del__(self):
        pass
        
    def __str__(self):
        return self.ClassPath()

class tkSimplexVariable(tkOptimizeData):
    def __init__(self, x0, dx, itmax = 100, tolx = 1.0e-5, tolf = 1.0e-5,
                callback = None, lprint = 1, iprintinterval = 10, 
                alpha = 1.0, beta = 0.5, gamma = 2.0, eps = 1.0e-50, **kwargs):
        self.nvar   = len(x0)
        self.nvtx   = self.nvar + 1
        self.x0     = x0
        self.x      = x0.copy()
        self.dx     = dx
        self.itmax  = itmax
        self.tolx   = tolx
        self.tolf   = tolf
        
        self.callback = callback
        self.lprint   = lprint
        self.iprintinterval = iprintinterval

        self.alpha  = alpha
        self.alpha1 = self.alpha + 1.0
        self.beta   = beta
        self.beta1  = 1.0 - self.beta
        self.gamma  = gamma
        self.gamma1 = 1.0 - self.gamma
        self.eps    = eps

        for i in range(self.nvar):
            if abs(self.dx[i]) < eps:
                self.dx[i] = abs(self.dx[i]) + 1.0

        self.vtx = []
        for i in range(self.nvtx):
            v = tkSimplexVertex(self.nvar)
            self.vtx.append(v)

        self.xr = []
        for i in range(self.nvar):
            self.xr.append(0.0)

        self.iter   = None
        self.il     = None
        self.ih     = None
        self.fl     = None
        self.fh     = None
        self.fs     = None
        self.fe     = None
        self.fr     = None
        self.fc     = None
        
        self.status = 0 # default is normal. status > 0 indicates an error
        
        self.update(**kwargs)

    def __del__(self):
        pass
        
    def __str__(self):
        return self.ClassPath()


def simplex(func = None, x0 = None, dx = None, itmax = 100, tolx = 1.0e-5, tolf = 1.0e-5, 
        optdata = None, callback = None, lprint = 1, iprintinterval = 10, 
        alpha = 1.0, beta = 0.5, gamma = 2.0,eps = 1.0e-50):
#       PARAMETER (MAXN=50, ALP=1.0, BET=0.5, GAM=2.0, EPS=1.D-50)
    """
        Function to perform simplex optimization
        Input:
            callback: Function called at every iteration
                    Return code should be 0 for continuation,
                                          >0 will terminate iteration
    """
    
#    print("x0=", x0)
    var = tkSimplexVariable(x0, dx, itmax, tolx, tolf, 
                callback, lprint, iprintinterval,
                alpha, beta, gamma, eps)
    if optdata is not None:
        optdata.nvtx   = var.nvtx
        optdata.vtx    = var.vtx
        optdata.xr     = var.xr
        optdata.alpha  = var.alpha
        optdata.alpha1 = var.alpha1
        optdata.beta   = var.beta
        optdata.beta1  = var.beta1
        optdata.gamma  = var.gamma
        optdata.gamma1 = var.gamma1
        var = optdata
    vtx = var.vtx
    x = x0.copy()
    
    nvar = len(x0)

#=============================================================
# SET INITIAL SIMPLEX
#=============================================================
    for j in range(nvar):
        vtx[0].x[j] = x[j]

    for i in range(nvar):
        vtx[i+1].x[i] = x[i] + dx[i]

    for i in range(0, var.nvtx):
#        print("i,x=", i, vtx[i].x)
        vtx[i].f = func(vtx[i].x, var)
#    exit()
        
    var.il = 0
    var.x0 = var.vtx[0].x
    var.x  = var.x0
    var.fl = func(var.x0, var)
    var.f  = var.fl

#=============================================================
#/* **  START OF THE MAIN LOOP */
#=============================================================
    for iter in range(0, itmax):
        var.iter = iter
        
        if callback is not None:
            var.x = var.vtx[var.il].x.copy()
            var.f = var.fl
            ret = callback(var)
            if ret > 0:
                var.status = ret
                return None, None, var

        if (iter-1) % iprintinterval == 0 and lprint <= 2:
            lvlprint(lprint, 1, "Simplex at iter {}:".format(iter))
            for i in range(var.nvtx):
                xstr = joinf(vtx[i].x, "%10.8g", ", ")
                lvlprint(lprint, 1, "  ", i, ": (", xstr, ")  f={:16.12g}".format(vtx[i].f))

#  FIND THE HIGHEST (IH, FH), THE SECOND HIGHEST (FS), */
#    AND THE LOWEST (IL, FL) POINTS */
        var.ih = 0
        var.fh = vtx[0].f
        var.il = 0
        var.fl = vtx[0].f
        for i in range(1, var.nvtx):
            if vtx[i].f > var.fh:
                var.ih = i
                var.fh = vtx[i].f
            if vtx[i].f < var.fl:
                var.il = i
                var.fl = vtx[i].f

        var.fs = var.fl
        for i in range(var.nvtx):
            if i != var.ih and vtx[i].f > var.fs:
                var.fs = vtx[i].f

        if (iter-1) % iprintinterval == 0:
            lvlprint(lprint, 2, 
                    "ITER={:3d}  FL, FS, FH={},{},{}".format(iter, var.fl, var.fs, var.fh))

# CONVERGENCE CHECK
#        IF(TOLX .GT. EPS) THEN
        if tolx > eps:
            xint = 0.0
#          DO 240 I1 = 1, N
            for i1 in range(nvar):
#          DO 240 I2 = I1+1, N+1
                for i2 in range(i1 + 1, var.nvtx):
#            DO 230 J = 1, N
                    for j in range(nvar):
#              XDIF = ABS( XX(J, I1) - XX(J, I2) )
                        xdif = abs(vtx[i1].x[j] - vtx[i2].x[j])
                        if xdif > xint:
                            xint = xdif
#  230       IF(XDIF .GT. XINT) XINT = XDIF
#  240     CONTINUE
#          IF(TOLF .GT. EPS) THEN
            if tolf > eps:
#            IF((FH-FL) .LT. TOLF .AND. XINT .LT. TOLX) GO TO 500
                if var.fh - var.fl < tolf and xint < tolx:
                    lvlprint(lprint, 0, "(SUBR.SMPLX)1 CONVERGENCE AT ITER={}".format(iter))
                    lvlprint(lprint, 0, "  fh={} - fl={} < tolf={}, xint={} < tolx={}"
                                .format(var.fh, var.fl, tolf, xint, tolx))
#      DO 510 J = 1, N
                    for j in range(nvar):
                        x[j] = vtx[var.il].x[j]
#  510 X(J) = XX(J, IL)
                    return x, vtx[var.il].f, var
#          ELSE
#            IF(XINT .LT. TOLX) GO TO 500
            elif xint < tolx:
#                goto .L500
                lvlprint(lprint, 0, "(SUBR.SMPLX)2 CONVERGENCE AT ITER={}".format(iter))
                return vtx[var.il].x, vtx[var.il].f, var
#          END IF
#        ELSE
        else:
#          IF(TOLF .GT. EPS .AND. (FH-FL) .LT. TOLF) GO TO 500
            if tolf > eps and fh - fl < tolf:
#            goto .L500
                lvlprint(lprint, 0, "(SUBR.SMPLX)3 CONVERGENCE AT ITER={}".format(iter))
                return vtx[var.il].x, vtx[var.il].f, var
#        END IF

# REFLECTION
#  X(*) IS THE CENTROID
#        DO 260 J = 1, N
        for j in range(nvar):
            s = 0.0
#          DO 250 I = 1, N+1
            for i in range(0, var.nvtx):
                s += vtx[i].x[j]
#  250     S = S + XX(J, I)
            x[j] = (s - vtx[var.ih].x[j]) / nvar
#  260   X(J) = (S - XX(J, IH)) / N

#  XR(*) IS THE NEW POINT BY REFELCTION */
#        DO 270 J = 1, N
        for j in range(nvar):
#            print("l194: j,ih={},{}".format(j,ih))
            var.xr[j] = var.alpha1 * x[j] - var.alpha * vtx[var.ih].x[j]
#  270   XR(J) = ALP1*X(J) - ALP*XX(J,IH)
        var.fr = func(var.xr, var)
        if (iter-1) % iprintinterval == 0:
            lvlprint(lprint, 2, "REFLECTION APPLIED.  FR ={}".format(var.fr), "  XR(*)=", var.xr)

# SEE WHERE FR LIES
        if var.fr <= var.fs:
            if var.fr < var.fl:
# EXPANSION    X(*) WILL BE XE IN THE NEXT LOOP
#            DO 280 J = 1, N
                for j in range(nvar):
                    x[j] = var.gamma * var.xr[j] + var.gamma1 * x[j]
#  280       X(J) = GAM*XR(J) + GAM1*X(J)
                var.fe = func(x, var)
                if (iter-1) % iprintinterval == 0:
                    lvlprint(lprint, 2, "EXPANSION APPLIED.   FE ={}".format(var.fe), "   XE(*)=", x)

                if var.fe < var.fr:
#              DO 290 J = 1, N
                    for j in range(nvar):
                       vtx[var.ih].x[j] = x[j]
#  290         XX(J, IH) = X(J)
                    vtx[var.ih].f = var.fe
                    if (iter-1) % iprintinterval == 0:
                        lvlprint(lprint, 2, "XH(*)1 IS REPLACED BY XE(*)")
                else:
#              DO 300 J = 1, N
                    for j in range(nvar):
                        vtx[var.ih].x[j] = var.xr[j]
#  300         XX(J, IH) = XR(J)
                vtx[var.ih].f = var.fr;
                if (iter-1) % iprintinterval == 0:
                    lvlprint(lprint, 2, "XH(*)2 IS REPLACED BY XR(*)")
            else:
#            DO 310 J = 1, N
                for j in range(nvar):
                    vtx[var.ih].x[j] = var.xr[j]
#  310       XX(J, IH) = XR(J)
                vtx[var.ih].f = var.fr
                if (iter-1) % iprintinterval == 0:
                    lvlprint(lprint, 2, "XH(*)3 IS REPLACED BY XR(*)")

# CONTRACTION    X(*) IS STILL THE CENTROID, BUT WILL SOON BE XC
        else:
            if var.fr < var.fh:
#           DO 320 J = 1, N
                for j in range(nvar):
                    vtx[var.ih].x[j] = var.xr[j]
#  320       XX(J, IH) = XR(J)
                vtx[var.ih].f = var.fr
                if (iter-1) % iprintinterval == 0:
                    lvlprint(lprint, 2, "XH(*)4 IS REPLACED BY XR(*)")
#          DO 330 J = 1, N
# some of vtx[].x might be identical to x[]. 
# To avoid this problem, x[] is first stored to a new list xtemp[]
            xtemp = np.empty(nvar)
            for j in range(nvar):
                xtemp[j] = var.beta * vtx[var.ih].x[j] + var.beta1 * x[j]
#                x[j] = var.beta * vtx[var.ih].x[j] + var.beta1 * x[j]
            x = xtemp
#  330     X(J) = BET*XX(J, IH) + BET1*X(J)
            var.fc = func(x, var)
            if (iter-1) % iprintinterval == 0:
                lvlprint(lprint, 2,"CONTRACTION APPLIED. FC ={}".format(var.fc), "  XC(*)=", x)
            if var.fc < var.fh:
#            DO 340 J = 1, N
                for j in range(nvar):
                    vtx[var.ih].x[j] = x[j]
#  340       XX(J, IH) = X(J)
                vtx[var.ih].f = var.fc
                if (iter-1) % iprintinterval == 0:
                    lvlprint(lprint, 2, "XH(*)5 IS REPLACED BY XC(*)")
            else:
# REDUCTION
#            IF(ITER .EQ. ITMAX) GO TO 401
                if iter == itmax:
#                goto .L401
                    lvlprint(lprint, 0, "(SUBR.SMPLX) ITERATION TERMINATED DUE TO ITMAX")
                    for j in range(nvar):
#  410 X(J) = XX(J, IL)
                        return vtx[var.il].x, vtx[var.il].f,var

#            DO 360 I = 1, N+1
                for i in range(var.nvtx):
#              DO 350 J = 1, N
# some of vtx[].x is identical to x.
# avoid this problem by reset x
                    x = np.empty(nvar)
                    for j in range(nvar):
#              X(J) = 0.5 * (XX(J, I) + XX(J, IL))
                        x[j] = 0.5 * (vtx[i].x[j] + vtx[var.il].x[j])
                    vtx[i].x = x
#  350         XX(J, I) = X(J)
                    vtx[i].f = func(x, var)
#  360       CONTINUE
                if (iter-1) % iprintinterval == 0:
                    lvlprint(lprint, 2, "REDUCTION APPLIED AROUND THE POINT: ", vtx[var.il].x)

    if (iter-1) % iprintinterval == 0:
        lvlprint(lprint, 0, "\nError: ITERATION TERMINATED DUE TO ITMAX")
    return vtx[var.il].x, vtx[var.il].f, var
