import openpyxl
from math import exp, sqrt, log
import numpy as np
from scipy import optimize          # newton関数はscipy.optimizeモジュールに入っている
from scipy.interpolate import interp1d


from tklib.tktransport.tkDOS_FEA_single_carrier import tkDOS_FEA_single_carrier
#from tkobject import tkObject
from tklib.tkutils import terminate, pfloat, pint, safe_getelement, validate_error, split_optstr
from tklib.tkinifile import tkIniFile
from tklib.tkvariousdata import tkVariousData
from tklib.tksci.tksci import pi, h, hbar,c, e, kB, me
from tklib.tkcrystal.tkvasp import tkVASP
from tklib.tktransport.tkTransport import read_datafile
from tklib.tktransport.tkTransport import LorentzNumber_FEA, FermiIntegral_x2
from tklib.tktransport.tkTransport import FermiIntegral_fast as FermiIntegral
from tklib.tktransport.tkTransport import rieman, integrate_Simpson, integrate_Simpson_list, fe, fh
from tklib.tktransport.tkTransport import meff2NC_FEA, NC2meff_FEA, meff2DC0_FEA


kpi05   = 2.0 / sqrt(pi)
kBe     = kB / e
kBe2    = kB * kB / e / e
ksigma0 = 16.0 * pi * e * e / 3.0 / h**3 * me
kmu0    = 2.0 * e / 3.0 / sqrt(2.0 * me)
ktau0   = 2.0 / 3.0 * sqrt(me * 0.5)
kS2     = kB / e * pi * pi / 3.0 * 1.5
kS3     = 8.0 / 3.0 * pi * pi * kB * kB / e / h / h * pow(pi / 3.0, 2.0/3.0) 
kkappa0 = 16.0 * pi * kB / 3.0 / h**3 * me


class tkDOS(tkDOS_FEA_single_carrier):
    def __init__(self):
        super(tkDOS_FEA_single_carrier, self).__init__()

        self.print_level    = 0
        
        self.mobility_e     = None
        self.mobility_h     = None
        self.cal_mobility_e = lambda T: 1.0
        self.cal_mobility_h = lambda T: 1.0
        self.cal_mobility_e_description = "Same mobility for hole and electron"
        self.cal_mobility_h_description = "Same mobility for hole and electron"

        self.meeff = None
        self.mheff = None
        self.NV    = None
        self.NC    = None
        self.DV0   = None
        self.DC0   = None

        self.EV    = None
        self.EC    = None
        self.Eg    = None

        self.NA    = None  # cm^-3
        self.ND    = None  # cm^-3
        self.EA    = None  # Acceptor states are not considered if EA = None
        self.ED    = None  # Donor states are not considered if ED = None

        self.EF0   = None
        self.varkeys  = None

# DOS list
        self.Elist    = None
        self.DOSlist  = None
        self.nE       = None
        self.Emin     = None
        self.Emax     = None
        self.Estep    = None

# EF range
        self.nEF      = None
        self.EFmin    = None
        self.EFmax    = None
        self.EFstep   = None

# integration parameters
# Integration range w.r.t. kBT
        self.Einteg0 = -3.0
        self.Einteg1 =  3.0
        self.nrange  =  6.0

# Bisection method parameters
# Initial search range, EPS, max iteration number, output cycle
        self.dE_bisec            = 0.4
        self.eps_bisec           = 1.0e-7
        self.nmaxiter_bisec      = 200
        self.iprintiterval_bisec = 1

        self.xT     = None
        self.yN     = None
        self.ymu    = None
        self.ysigma = None
        self.yRH    = None

    def __del__(self):
#        super(tkObject, self).__del__()
        pass

    def __str__(self):
        return self.ClassPath()


    def cal_carrier_densities(self, T, EF, E0h, E1h, E0e, E1e, Estep):
#        print("T={:8.4g} EF={:8.4g}   EV={:8.4g} E range h: {:8.4g} - {:8.4g}".format(T, EF, self.EV, E0h, E1h))
#        print("                       EC={:8.4g} E range h: {:8.4g} - {:8.4g}".format(self.EC, E0e, E1e))
        Nh  = self.Nh(T, EF, E0h, E1h, Estep)
        Ne  = self.Ne(T, EF, E0e, E1e, Estep)
        NAm = self.NAm(EF, T)
        NDp = self.NDp(EF, T)
        RH  = 1.0 / e * (Nh - Ne) / pow(Nh + Ne + 1.0e-50, 2.0)
        Ns  = 1.0 / e / (RH + 1.0e-300)
        
        return Nh, Ne, NAm, NDp, RH, Ns

    def cal_RH(self, muh, nh, mue, ne):
        return (nh * muh * muh - ne * mue * mue) / pow(nh * muh + ne * mue, 2.0) / e

# Hall-related properties for two carrier model
    def cal_transport_properties_single_carrier(self, x, T, me, NC, r, l0norm):
        ksigma = ksigma0 * l0norm * me
        kmu    = kmu0 * l0norm / sqrt(me) 
        ktau   = ktau0 * l0norm * (r + 1.0) * sqrt(me) 

        Fr      = FermiIntegral(x, r)
        Fr1     = FermiIntegral(x, r + 1.0)
        Fr2     = FermiIntegral(x, r + 2.0)
        F05     = FermiIntegral(x, 0.5)
        F2rhalf = FermiIntegral(x, 2.0 * r - 0.5)
        
        n      = NC * kpi05 * F05                                  # cm^-3
        sigma  = ksigma * pow(kB * T, r + 1.0) * (r + 1.0) * Fr / 100.0 # S/cm
        mu     = sigma / e / n                                          # cm2/Vs
        FH     = 1.5 * (2.0 * r + 0.5) / (r + 1.0)**2 * F05 * F2rhalf / Fr / Fr
        RH0    = 1.0 / e / n
        RH     = FH * RH0
        nHall  = n / FH
        muHall = FH * mu
        
        return sigma, n, mu, FH, RH0, RH, nHall, muHall
    
    def cal_Hall_properteis2(self, T, EF, mobility, validate_error_str = None):
        mobility.set_scattering_parameters()
        self.NC  = meff2NC_FEA(self.meeff, T)
        self.DC0 = meff2DC0_FEA(self.meeff, T)

        kBT = kB * T
        kT = e / kBT
        xe  = (EF - self.EC) * kT
        xh  = (self.EV - EF) * kT
        sigmae, ne, mue, FHe, RH0e, RHe, nHalle, muHalle = self.cal_transport_properties_single_carrier(xe, T, self.meeff, self.NC, mobility.rfac, mobility.l0norm)
        sigmah, nh, muh, FHh, RH0h, RHh, nHallh, muHallh = self.cal_transport_properties_single_carrier(xh, T, self.mheff, self.NV, mobility.rfac, mobility.l0norm)

        NAm = self.NAm(EF, T)
        NDp = self.NDp(EF, T)
        dQ  = ne + NAm - nh - NDp

        if self.cal_mobility_h is None:
            muh_RH = muh
        else:
            muh_RH = self.cal_mobility_e(T)
        if self.cal_mobility_e is None:
            mue_RH = mue
        else:
            mue_RH = self.cal_mobility_e(T)
#        RH  = 1.0 / e * (nHallh - nHalle) / pow(nHallh + nHalle, 2.0)
        RH  = self.cal_RH(muh_RH, nHallh, mue_RH, nHalle)
        Ns  = 1.0 / e / RH
        
        inf = {
              "sigma_e": sigmae, 
              "sigma_h": sigmah, 
              "mu_e" : mue, "muHall_e": muHalle, #"<tau_e>": tau_avg_e,
              "mu_h" : muh, "muHall_h": muHallh, #"<tau_h>": tau_avg_h,
              "n_e"  : ne,  "nHall_e" : nHalle, 
              "n_h"  : nh,  "nHall_h" : nHallh, 
              "RH": RH, "n_s": Ns,
              "FH_e": FHe, "RH_e": RHe, "RH0_e": RH0e,
              "FH_h": FHh, "RH_h": RHh, "RH0_h": RH0h,
              "NAm": NAm, "NDp": NDp, "dQ": dQ
            }
            
        return inf

    def cal_T_lists2(self, xT, mobility, print_level = 0):
        xTinv   = []
        yEF     = []
        yNe     = []
        yNh     = []
        yNeHall = []
        yNhHall = []
        yRHe    = []
        yRHh    = []
        yFHalle = []
        yFHallh = []
        yNDp    = []
        yNAm    = []
        yRH     = []
        yNs     = []
        yNsabs  = []
        ydQ     = []
        EFprev  = self.EF0
#初期範囲
        EFmin = min([self.EV, self.EF0 - self.dE_bisec])
        EFmax = max([self.EC, self.EF0 + self.dE_bisec])
#    dEFbisec = 0.1
        if print_level >= 1:
            print("")
        for iT in range(len(xT)):
            T = xT[iT]

            if print_level >= 2:
                print("iT {:03d}: {} K".format(iT, T))

            kBTe = kB * T / e
            dEinteg = self.nrange * kBTe

            E0h = self.EV - dEinteg
            E1e = self.EC + dEinteg

#            if E0h < self.Emin or self.Emax < E1e:
#                if print_level >= 1:
#                    print("*** Integration range error: [{}, {}] exceeds DOS E range [{}, {}]".format(E0h, E1e, self.Emin, self.Emax))
#                exit()

            self.NC  = self.meff2NC_FEA(self.meeff, T)
            self.NV  = self.meff2NC_FEA(self.mheff, T)
            self.DC0 = self.meff2DC0_FEA(self.meeff, T)
            self.DV0 = self.meff2DC0_FEA(self.mheff, T)

#        Ne0K = self.Ne(0.0, self.EF0, E0h, self.EF0, self.Estep)
#            print("EF range:", EFmin, EFmax)
            EF, dQ = self.FindEF_semi(T, EFmin, EFmax, E0h, self.EV, self.EC, E1e, self.Estep)
#            Nhh, Neh, NAmh, NDph, RH, Ns = self.cal_carrier_densities(T, EF, E0h, self.EV, self.EC, E1e, self.Estep)
            inf  = self.cal_Hall_properteis2(T, EF, mobility, validate_error_str = None)

            xTinv.append(1000.0 / T)
            yEF.append(EF)
            yNe.append(inf['n_e'])
            yNh.append(inf['n_h'])
            yFHalle.append(inf['FH_e'])
            yFHallh.append(inf['FH_h'])
            yRHe.append(inf['RH_e'])
            yRHh.append(inf['RH_h'])
            yNeHall.append(inf['nHall_e'])
            yNhHall.append(inf['nHall_h'])
            yNDp.append(inf['NDp'])
            yNAm.append(inf['NAm'])
            yRH.append(inf['RH'])
            yNs.append(inf['n_s'])
            yNsabs.append(abs(inf['n_s']))
            ydQ.append(dQ)
            if print_level >= 2:
                print("  ***(T,EF,Ne,Nh,NDp,NAm,RH)=({:6.4f},{:6.4g},{:10.4g},{:10.4g},{:10.4g},{:10.4g},{:10.4g},{:10.4g})"
                            .format(T, EF, inf['n_h'], inf['n_e'], inf['NDp'], inf['NAm'], inf['RH'], inf['n_s']))

            EFmin = min([self.EV, EF - self.dE_bisec])
            EFmax = max([self.EC, EF + self.dE_bisec])

        inf = { "T": xT, "Tinv": xTinv,
                "EF": yEF, "dQ": ydQ, "RH": yRH, 'n_s': yNs, 'n_s(abs)': yNsabs,
                "n_e": yNe, "nHall_e": yNeHall, 'FHall_e': yFHalle, 'RH_e': yRHe,
                "n_h": yNh, "nHall_h": yNhHall, 'FHall_h': yFHallh, 'RH_h': yRHh,
                "NAm": yNAm, "NDp": yNDp
              }

        return inf    
    
def main():
    print("")
    print("This is a library, not runnable")
    print("")


if __name__ == "__main__":
    main()
